from __future__ import annotations

import uuid
from typing import Any
from unittest import mock

import pytest

from academy.exchange.cloud.authenticate import get_authenticator
from academy.exchange.cloud.authenticate import get_token_from_headers
from academy.exchange.cloud.authenticate import GlobusAuthenticator
from academy.exchange.cloud.authenticate import NullAuthenticator
from academy.exchange.cloud.config import ExchangeAuthConfig
from academy.exchange.cloud.exceptions import ForbiddenError
from academy.exchange.cloud.exceptions import UnauthorizedError


def test_null_authenticator() -> None:
    user1 = NullAuthenticator().authenticate_user({})
    user2 = NullAuthenticator().authenticate_user({'Authorization': 'token'})
    assert user1 == user2


def test_authenticate_user_with_token() -> None:
    authenticator = GlobusAuthenticator(str(uuid.uuid4()), '')

    token_meta: dict[str, Any] = {
        'active': True,
        'aud': [authenticator.audience],
        'sub': authenticator.auth_client.client_id,
        'username': 'username',
        'client_id': str(uuid.uuid4()),
        'email': 'username@example.com',
        'name': 'User Name',
    }

    with mock.patch.object(
        authenticator.auth_client,
        'oauth2_token_introspect',
        return_value=token_meta,
    ):
        user = authenticator.authenticate_user(
            {'Authorization': 'Bearer <TOKEN>'},
        )

    assert user == uuid.UUID(token_meta['client_id'])


def test_authenticate_user_with_token_expired_token() -> None:
    authenticator = GlobusAuthenticator(str(uuid.uuid4()), '')
    with (
        mock.patch.object(
            authenticator.auth_client,
            'oauth2_token_introspect',
            return_value={'active': False},
        ),
        pytest.raises(
            ForbiddenError,
            match='Token is expired or has been revoked.',
        ),
    ):
        authenticator.authenticate_user({'Authorization': 'Bearer <TOKEN>'})


def test_authenticate_user_with_token_wrong_audience() -> None:
    authenticator = GlobusAuthenticator(
        str(uuid.uuid4()),
        '',
        audience='audience',
    )
    with (
        mock.patch.object(
            authenticator.auth_client,
            'oauth2_token_introspect',
            return_value={'active': True},
        ),
        pytest.raises(
            ForbiddenError,
            match='Token audience does not include "audience"',
        ),
    ):
        authenticator.authenticate_user({'Authorization': 'Bearer <TOKEN>'})


def test_get_authenticator() -> None:
    config = ExchangeAuthConfig()
    authenticator = get_authenticator(config)
    assert isinstance(authenticator, NullAuthenticator)

    config = ExchangeAuthConfig(
        method='globus',
        kwargs={
            'audience': 'test',
            'client_id': str(uuid.uuid4()),
            'client_secret': 'test',
        },
    )
    authenticator = get_authenticator(config)
    assert isinstance(authenticator, GlobusAuthenticator)
    assert authenticator.audience == 'test'


def test_get_authenticator_unknown() -> None:
    config = ExchangeAuthConfig(method='globus')
    # Modify attribute after construction to avoid Pydantic checking string
    # literal type.
    config.method = 'test'  # type: ignore[assignment]
    with pytest.raises(ValueError, match='Unknown authentication method'):
        get_authenticator(config)


def test_get_token_from_headers() -> None:
    headers = {'Authorization': 'Bearer <TOKEN>'}
    assert get_token_from_headers(headers) == '<TOKEN>'


def test_get_token_from_headers_missing() -> None:
    with pytest.raises(
        UnauthorizedError,
        match='Request headers are missing authorization header.',
    ):
        get_token_from_headers({})


def test_get_token_from_headers_malformed() -> None:
    with pytest.raises(
        UnauthorizedError,
        match='Bearer token in authorization header is malformed.',
    ):
        get_token_from_headers({'Authorization': '<TOKEN>'})



from __future__ import annotations

import pickle
from typing import Any


class NoPickleMixin:
    """Mixin that raises an error if a type is pickled."""

    def __getstate__(self) -> Any:
        raise pickle.PicklingError(f'{type(self).__name__} is not pickleable.')



from __future__ import annotations

import pickle
from collections.abc import AsyncGenerator
from typing import Any

import pytest
import pytest_asyncio

from academy.agent import Agent
from academy.exception import BadEntityIdError
from academy.exception import MailboxTerminatedError
from academy.exchange import ExchangeFactory
from academy.exchange import MailboxStatus
from academy.exchange.hybrid import HybridExchangeFactory
from academy.exchange.transport import AgentRegistrationT
from academy.exchange.transport import ExchangeTransport
from academy.identifier import AgentId
from academy.identifier import UserId
from academy.message import ErrorResponse
from academy.message import Message
from academy.message import PingRequest
from academy.message import SuccessResponse
from testing.agents import EmptyAgent
from testing.constant import TEST_SLEEP_INTERVAL
from testing.fixture import EXCHANGE_FACTORY_TYPES


@pytest_asyncio.fixture(params=EXCHANGE_FACTORY_TYPES)
async def transport(
    request,
    get_factory,
) -> AsyncGenerator[ExchangeTransport[AgentRegistrationT]]:
    factory = get_factory(request.param)
    async with await factory._create_transport() as transport:
        yield transport


@pytest.mark.asyncio
async def test_transport_repr(
    transport: ExchangeTransport[AgentRegistrationT],
) -> None:
    assert isinstance(repr(transport), str)
    assert isinstance(str(transport), str)


@pytest.mark.asyncio
async def test_transport_create_factory(
    transport: ExchangeTransport[AgentRegistrationT],
) -> None:
    new_factory = transport.factory()
    assert isinstance(new_factory, ExchangeFactory)


@pytest.mark.asyncio
async def test_transport_register_agent(
    transport: ExchangeTransport[AgentRegistrationT],
) -> None:
    registration = await transport.register_agent(EmptyAgent)
    status = await transport.status(registration.agent_id)
    assert status == MailboxStatus.ACTIVE


@pytest.mark.asyncio
async def test_transport_status(
    transport: ExchangeTransport[AgentRegistrationT],
) -> None:
    uid = UserId.new()
    status = await transport.status(uid)
    assert status == MailboxStatus.MISSING
    registration = await transport.register_agent(EmptyAgent)
    status = await transport.status(registration.agent_id)
    assert status == MailboxStatus.ACTIVE
    await transport.terminate(registration.agent_id)
    await transport.terminate(registration.agent_id)  # Idempotency
    status = await transport.status(registration.agent_id)
    assert status == MailboxStatus.TERMINATED


@pytest.mark.asyncio
async def test_transport_send_recv(
    transport: ExchangeTransport[AgentRegistrationT],
) -> None:
    for _ in range(3):
        message = Message.create(
            src=transport.mailbox_id,
            dest=transport.mailbox_id,
            body=PingRequest(),
        )
        await transport.send(message)
        assert await transport.recv() == message


@pytest.mark.asyncio
async def test_transport_send_bad_identifier_error(
    transport: ExchangeTransport[AgentRegistrationT],
) -> None:
    uid: AgentId[Any] = AgentId.new()
    with pytest.raises(BadEntityIdError):
        await transport.send(
            Message.create(
                src=transport.mailbox_id,
                dest=uid,
                body=PingRequest(),
            ),
        )


@pytest.mark.asyncio
async def test_transport_send_mailbox_closed(
    transport: ExchangeTransport[AgentRegistrationT],
) -> None:
    registration = await transport.register_agent(EmptyAgent)
    await transport.terminate(registration.agent_id)
    with pytest.raises(MailboxTerminatedError):
        await transport.send(
            Message.create(
                src=transport.mailbox_id,
                dest=registration.agent_id,
                body=PingRequest(),
            ),
        )


@pytest.mark.asyncio
async def test_transport_recv_mailbox_closed(
    transport: ExchangeTransport[AgentRegistrationT],
) -> None:
    await transport.terminate(transport.mailbox_id)
    with pytest.raises(MailboxTerminatedError):
        await transport.recv()


@pytest.mark.asyncio
async def test_transport_recv_timeout(
    transport: ExchangeTransport[AgentRegistrationT],
) -> None:
    with pytest.raises(TimeoutError):
        await transport.recv(timeout=TEST_SLEEP_INTERVAL)


@pytest.mark.asyncio
async def test_transport_terminate_unknown_ok(
    transport: ExchangeTransport[AgentRegistrationT],
) -> None:
    await transport.terminate(UserId.new())


@pytest.mark.parametrize('factory_type', EXCHANGE_FACTORY_TYPES)
async def test_transport_terminate_reply_pending_requests(
    factory_type: type[ExchangeFactory[Any]],
    get_factory,
) -> None:
    if factory_type is HybridExchangeFactory:
        pytest.skip(
            'HybridExchangeTransport termination behavior is unreliable.',
        )

    factory = get_factory(factory_type)
    async with await factory._create_transport() as transport1:
        async with await factory._create_transport() as transport2:
            # Put a request and a response message in transport2 mailbox
            message1 = Message.create(
                src=transport1.mailbox_id,
                dest=transport2.mailbox_id,
                body=PingRequest(),
            )
            message2 = Message.create(
                src=transport1.mailbox_id,
                dest=transport2.mailbox_id,
                body=SuccessResponse(),
            )
            await transport1.send(message1)
            await transport1.send(message2)

            # Terminate transport2 mailbox should reply to all *requests*
            # with an error. Responses are ignored.
            await transport2.terminate(transport2.mailbox_id)

            # Check that transport1 gets a response to its request that
            # was terminated.
            response = await transport1.recv()
            body = response.get_body()
            assert isinstance(body, ErrorResponse)
            assert response.tag == message1.tag
            assert isinstance(body.get_exception(), MailboxTerminatedError)

            # No other messages should have been received
            with pytest.raises(TimeoutError):
                await transport1.recv(timeout=TEST_SLEEP_INTERVAL)


@pytest.mark.asyncio
async def test_transport_non_pickleable(
    transport: ExchangeTransport[AgentRegistrationT],
) -> None:
    with pytest.raises(pickle.PicklingError):
        pickle.dumps(transport)


class A(Agent): ...


class B(Agent): ...


class C(B): ...


@pytest.mark.asyncio
async def test_transport_discover(
    transport: ExchangeTransport[AgentRegistrationT],
) -> None:
    bid = (await transport.register_agent(B)).agent_id
    cid = (await transport.register_agent(C)).agent_id
    did = (await transport.register_agent(C)).agent_id
    await transport.terminate(did)

    found = await transport.discover(A)
    assert len(found) == 0
    found = await transport.discover(B, allow_subclasses=False)
    assert found == (bid,)
    found = await transport.discover(B, allow_subclasses=True)
    assert found == (bid, cid)



from __future__ import annotations

import pytest

# Mark all tests in tests/integration as integration tests so they
# can be selected or deselected with:
#   pytest -k "integration"
#   pytest -k "not integration"
pytestmark = pytest.mark.integration
